package main
import (
	"fmt"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"strconv"
	"strings"
)
// HeroesServiceChaincode implementation of Chaincode
type HeroesServiceChaincode struct {
}
// Init of the chaincode
// This function is called only one when the chaincode is instantiated.
// So the goal is to prepare the ledger to handle future requests.
func (t *HeroesServiceChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	fmt.Println("######## HeroesServiceChaincode Init ########")
	// Get the function and arguments from the request
	function, _ := stub.GetFunctionAndParameters()
	// Check if the request is the init function
	if function != "init" {
		return shim.Error("Unknown function call")
	}
	//Put in the ledger the key/value hello/world
	err := stub.PutState("hello", []byte("world"))
	if err != nil {
		return shim.Error(err.Error())
	}
	err = stub.PutState("schedules", []byte("0"))
	if err != nil {
		return shim.Error(err.Error())
	}
	// Return a succesful message
	return shim.Success(nil)
}
// Invoke
// All future requests named invoke will arrive here
func (t *HeroesServiceChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	fmt.Println("######## HeroesServiceChaincode Invoke ########")
	// Get the function and arguments from the request
	function, args := stub.GetFunctionAndParameters()
	// Check whether it is an invoke request
	if function != "invoke" {
		return shim.Error("Unknown functin call")
	}
	// Check whether the number of arguments is sufficient
	if len(args) < 1 {
		return shim.Error("The number of arguments is insufficient.")
	}
	// In order to manage multiple type of request, we will check the first argument.
	// Here we have one possible argument: query (every query request will read
	// in the ledger without modification)
	if args[0] == "query" {
		return t.query(stub, args)
	}
	// Add for ledger state changing. 
	// The update argument will manage all update in the ledger
	if args[0] == "invoke" {
		return t.invoke(stub, args)
	}
	if args[0] == "addOperation" {
		return t.addOperation(stub, args)
	}
	if args[0] == "addPki" {
		return t.addPki(stub, args)
	}
	if args[0] == "addResource" {
		return t.addResource(stub, args)
	}
	// Query.
	if args[0] == "queryOperation" {
		return t.queryOperation(stub, args)
	}
	if args[0] == "queryPki" {
		return t.queryPki(stub, args)
	}
	if args[0] == "queryResource" {
		return t.queryResource(stub, args)
	}
	if args[0] == "querySchedulesIDs" {
		return t.querySchedulesIDs(stub, args)
	}
	if args[0] == "queryProcessesIDsOfSchedule" {
		return t.queryProcessesIDsOfSchedule(stub, args)
	}
	if args[0] == "queryOperationsIDsOfProcess" {
		return t.queryOperationsIDsOfProcess(stub, args)
	}
	if args[0] == "queryPkisIDsOfSchedule" {
		return t.queryPkisIDsOfSchedule(stub, args)
	}
	if args[0] == "queryResourcesIDsOfSchedule" {
		return t.queryResourcesIDsOfSchedule(stub, args)
	}
	return shim.Error("Unknown action, check the first argument")
}
// query
// Every readonly function in the ledger will be here
func (t *HeroesServiceChaincode) query(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## HeroesServiceChaincode query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 2 {
		return shim.Error("The number of arguments is insufficient.")
	}
	// Like the invoke function, we manage multiple type of query requests
	// with the second argument.
	// We also have only one possible argument: hello
	if args[1] == "hello" {
		// Get the state of the value matching the key hello in the ledger
		state, err := stub.GetState("hello")
		if err != nil {
			return shim.Error("Failed to get state of hello")
		}
		// Return this value
		return shim.Success(state)
	}
	// If the argument given don't match any function, we return an error
	return shim.Error("Unknown query action, check the second argument.")
}
// queryOperation. 
func (t *HeroesServiceChaincode) queryOperation(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot state query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 5 {
		return shim.Error("The number of arguments is insufficient.")
	}
	stateName		:=	args[1]
	botId			:=	args[2]
	procId			:=	args[3]
	operId			:=	args[4]
	// Key
	keySchedule := "schedule" + botId
	keyProcess	:= "proc" + procId
	keyOper		:= "oper" + operId
	keyComplex	:= keySchedule + "_" + keyProcess + "_" + keyOper + "_" + stateName
	// Like the invoke function, we manage multiple type of query requests
	// with the third argument.
	// Note possible arguments: operName, priority, resourceLink, startTime, endTime
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get state of " + stateName)
	}
	// Return this value
	return shim.Success(state) 
}
func (t *HeroesServiceChaincode) queryPki(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot state query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 4 {
		return shim.Error("The number of arguments is insufficient.")
	}
	stateName		:=	args[1]
	botId			:=	args[2]
	pkiId			:=	args[3]
	// Key
	keySchedule := "schedule" + botId
	keyPki		:= "pki" + pkiId
	keyComplex	:= keySchedule + "_" + keyPki + "_" + stateName
	// Like the invoke function, we manage multiple type of query requests
	// with the third argument.
	// Note possible arguments: pkiName, ...
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get state of " + stateName)
	}
	// Return this value
	return shim.Success(state) 
	//return shim.Error("Unknown query action, check the second argument.")
}
func (t *HeroesServiceChaincode) queryResource(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot state query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 4 {
		return shim.Error("The number of arguments is insufficient.")
	}
	stateName		:=	args[1]
	botId			:=	args[2]
	resId			:=	args[3]
	// Key
	keySchedule := "schedule" + botId
	keyRes		:= "res" + resId
	keyComplex	:= keySchedule + "_" + keyRes + "_" + stateName
	// Like the invoke function, we manage multiple type of query requests
	// with the third argument.
	// Note possible arguments: pkiName, ...
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get state of " + stateName)
	}
	// Return this value
	return shim.Success(state) 
}
// queryProcessesIDsOfSchedule. 
func (t *HeroesServiceChaincode) querySchedulesIDs(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot IDs query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 1 {
		return shim.Error("The number of arguments is insufficient.")
	}
	// Key
	keyComplex	:= "schedule_botIDs"
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get schedules list")
	}
	// Return this value
	return shim.Success(state)
}
// queryProcessesIDsOfSchedule. 
func (t *HeroesServiceChaincode) queryProcessesIDsOfSchedule(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot processes IDs query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 2 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId		:=	args[1]
	// Key
	keySchedule := "schedule" + botId
	keyComplex	:= keySchedule + "_procIDs"
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get processes list")
	}
	// Return this value
	return shim.Success(state)
}
// queryOperationsIDsOfProcess. 
func (t *HeroesServiceChaincode) queryOperationsIDsOfProcess(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot operations IDs query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 3 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId		:=	args[1]
	procId		:=	args[2]
	// Key
	keySchedule := "schedule" + botId
	keyProcess	:= "proc" + procId
	keyComplex	:= keySchedule + "_" + keyProcess + "_operIDs"
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get operations list")
	}
	// Return this value
	return shim.Success(state)
}
// queryPkisIDsOfSchedule. 
func (t *HeroesServiceChaincode) queryPkisIDsOfSchedule(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot pki IDs query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 2 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId		:=	args[1]
	// Key
	keySchedule := "schedule" + botId
	keyComplex	:= keySchedule + "_pkiIDs"
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get processes list")
	}
	// Return this value
	return shim.Success(state)
}
// queryResourcesIDsOfSchedule. 
func (t *HeroesServiceChaincode) queryResourcesIDsOfSchedule(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot resources IDs query ########")
	// Check whether the number of arguments is sufficient
	if len(args) < 2 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId		:=	args[1]
	// Key
	keySchedule := "schedule" + botId
	keyComplex	:= keySchedule + "_resIDs"
	state, err := stub.GetState(keyComplex)
	if err != nil {
		return shim.Error("Failed to get processes list")
	}
	// Return this value
	return shim.Success(state)
}
// Add for ledger state changing.
// invoke
// Every functions that read and write in the ledger wil be here
func (t *HeroesServiceChaincode) invoke(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## HeroesServiceChaincode invoke ########")
	if len(args) < 2 {
		return shim.Error("The number of arguments is insufficient.")
	}
	// Check if the ledger key is "hello" and process if it is the case. Otherwise
	// it returns an error
		if args[1] == "hello" && len(args) == 3 {
			// Write the new value in the ledger
			err := stub.PutState("hello", []byte(args[2]))
			if err != nil {
				return shim.Error("Failed to update state of hello")
			}
			// Notify listener that an event "eventInvoke" have been executed
			// (check line 19 in the file invoke.go)
			err = stub.SetEvent("eventInvoke", []byte{})
			if err != nil {
				return shim.Error(err.Error())
			}
			// Return this value in response
			return shim.Success(nil)
		}
	// If the argument given don't match any function, we return an error
	return shim.Error("Unknown invoke action, check the second argument.")
}
// Add for operation addition. 
// addOperation
// Every functions that read and write in the ledger will be here
func (t *HeroesServiceChaincode) addOperation(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot addOperation ########")
	if len(args) < 9 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId			:=	args[1]
	procId			:=	args[2]
	operId			:=	args[3]
	operName		:=	args[4]
	priority		:=	args[5]
	resourceLink	:=	args[6]
	startTimeAS		:=	args[7]
	endTimeAS		:=	args[8]
	// Key
	keySchedule := "schedule" + botId
	keyProcess	:= "proc" + procId
	keyOper		:= "oper" + operId
	keyComplex	:= keySchedule + "_" + keyProcess + "_" + keyOper
	// Counters Init
	countSchedules	:= 0
	countProcs	:= 0
	countOpers	:= 0
	// Counters and links
	bytes, err := stub.GetState("schedules")  // see Init()
	if err != nil {
		return shim.Error("Failed to get schedules count")
	}
	countSchedules, err = btoi(bytes)
	if err != nil {
		return shim.Error("Can't convert schedules count to int")
	}
	if countSchedules != 0 {
		bytes, err = stub.GetState(keySchedule + "_processes")
		if err != nil {
			return shim.Error("Failed to get processes count for schedule " + botId)
		}
		countProcs, err = btoi(bytes)
		if err != nil {
			return shim.Error("Can't convert processes count to int")
		}
	}
	if countProcs != 0 {
		bytes, err = stub.GetState(keySchedule + "_" + keyProcess + "_operations")
		if err != nil {
			return shim.Error("Failed to get operations count for " + keySchedule + "_" + keyProcess)
		}
		countOpers, err = btoi(bytes)
		if err != nil {
			return shim.Error("Can't convert operations count to int")
		}
	}
	// Putting the state...
	err = stub.PutState(keyComplex + "_operName", []byte(operName))
	if err != nil {
		return shim.Error("Failed to set the name of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_priority", []byte(priority))
	if err != nil {
		return shim.Error("Failed to set the priority of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_resourceLink", []byte(resourceLink))
	if err != nil {
		return shim.Error("Failed to set the resource of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_startTime", []byte(startTimeAS))
	if err != nil {
		return shim.Error("Failed to set the start time array of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_endTime", []byte(endTimeAS))
	if err != nil {
		return shim.Error("Failed to set the end time array of " + keyComplex)
	}
	// Increments
	up, err := t.updateList(stub, "schedule_botIDs", botId)
	if err != nil {
		return shim.Error("Failed to update the schedules list: " + err.Error())
	}
	if up {
		countSchedules++
		err = stub.PutState("schedules", itob(countSchedules))
		if err != nil {
			return shim.Error("Failed to update schedules count")
		}
	}
	up, err = t.updateList(stub, keySchedule + "_procIDs", procId)
	if err != nil {
		return shim.Error("Failed to update the processes list")
	}
	if up {
		countProcs++
		err = stub.PutState(keySchedule + "_processes", itob(countProcs))
		if err != nil {
			return shim.Error("Failed to update processes count")
		}
	}
	up, err = t.updateList(stub, keySchedule + "_" + keyProcess + "_operIDs", operId)
	if err != nil {
		return shim.Error("Failed to update the operations list")
	}
	if up {
		countOpers++
		err = stub.PutState(keySchedule + "_" + keyProcess + "_operations", itob(countOpers))
		if err != nil {
			return shim.Error("Failed to update operations count")
		}
	}
	// It's success
	err = stub.SetEvent("eventInvoke", []byte{})
	if err != nil {
		return shim.Error(err.Error())
	}
	// Return this value in response
	return shim.Success(nil)
}
func btoi(bs []byte) (int, error) {
	if len(bs) == 0 {
	    return 0, nil
	}
	sob := string(bs)
	return strconv.Atoi(sob)
}
func itob(cnt int) []byte {
	soi := strconv.Itoa(cnt)
	return []byte(soi)
}
// addPki
// Every functions that read and write in the ledger will be here
func (t *HeroesServiceChaincode) addPki(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot addPki ########")
	if len(args) < 7 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId			:=	args[1]
	pkiId			:=	args[2]
	generalized		:=	args[3]
	weight			:=	args[4]
	name			:=	args[5]
	value			:=	args[6]
	// Key
	keySchedule := "schedule" + botId
	keyPki		:= "pki" + pkiId
	keyComplex	:= keySchedule + "_" + keyPki
	// Counters Init
	countSchedules	:= 0
	countPkis	:= 0
	// Counters and links
	bytes, err := stub.GetState("schedules")  // see Init()
	if err != nil {
		return shim.Error("Failed to get schedules count")
	}
	countSchedules, err = btoi(bytes)
	if err != nil {
		return shim.Error("Can't convert schedules count to int")
	}
	if countSchedules != 0 {
		bytes, err = stub.GetState(keySchedule + "_pkis")
		if err != nil {
			return shim.Error("Failed to get PKIs count for schedule " + botId)
		}
		countPkis, err = btoi(bytes)
		if err != nil {
			return shim.Error("Can't convert PKIs count to int")
		}
	}
	// Putting of the state...
	err = stub.PutState(keyComplex + "_pkiName", []byte(name))
	if err != nil {
		return shim.Error("Failed to set the name of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_pkiWeight", []byte(weight))
	if err != nil {
		return shim.Error("Failed to set the weight of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_pkiGeneralized", []byte(generalized))
	if err != nil {
		return shim.Error("Failed to set the generalized state of " + keyComplex)
	}
	err = stub.PutState(keyComplex + "_pkiValue", []byte(value))
	if err != nil {
		return shim.Error("Failed to set the value of " + keyComplex)
	}
	// Increments
	up, err := t.updateList(stub, "schedule_botIDs", botId)
	if err != nil {
		return shim.Error("Failed to udpate the schedules list")
	}
	if up {
		countSchedules++
		err = stub.PutState("schedules", itob(countSchedules))
		if err != nil {
			return shim.Error("Failed to update schedules count")
		}
	}
	up, err = t.updateList(stub, keySchedule + "_pkiIDs", pkiId)
	if err != nil {
		return shim.Error("Failed to udpate the quality list")
	}
	if up {
		countPkis++
		err = stub.PutState(keySchedule + "_pkis", itob(countPkis))
		if err != nil {
			return shim.Error("Failed to update PKIs count")
		}
	}
	// It's success
	err = stub.SetEvent("eventInvoke", []byte{})
	if err != nil {
		return shim.Error(err.Error())
	}
	// Return this value in response
	return shim.Success(nil)
}
// addResource
// Every functions that read and write in the ledger will be here
func (t *HeroesServiceChaincode) addResource(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	fmt.Println("######## BorisSokolChaincode bot addResource ########")
	if len(args) < 4 {
		return shim.Error("The number of arguments is insufficient.")
	}
	botId			:=	args[1]
	resId			:=	args[2]
	worktime		:=	args[3]
	// Key
	keySchedule := "schedule" + botId
	keyRes		:= "res" + resId
	keyComplex	:= keySchedule + "_" + keyRes
	// Counters Init
	countSchedules	:= 0
	countResources	:= 0
	// Counters and links
	bytes, err := stub.GetState("schedules")  // see Init()
	if err != nil {
		return shim.Error("Failed to get schedules count")
	}
	countSchedules, err = btoi(bytes)
	if err != nil {
		return shim.Error("Can't convert schedules count to int")
	}
	if countSchedules != 0 {
		bytes, err = stub.GetState(keySchedule + "_resources")
		if err != nil {
			return shim.Error("Failed to get resources count for schedule " + botId)
		}
		countResources, err = btoi(bytes)
		if err != nil {
			return shim.Error("Can't convert resources count to int")
		}
	}
	// Putting of the state...
	err = stub.PutState(keyComplex + "_resWorktime", []byte(worktime))
	if err != nil {
		return shim.Error("Failed to set the worktime of " + keyComplex)
	}
	// Increments
	up, err := t.updateList(stub, "schedule_botIDs", botId)
	if err != nil {
		return shim.Error("Failed to udpate the schedules list")
	}
	if up {
		countSchedules++
		err = stub.PutState("schedules", itob(countSchedules))
		if err != nil {
			return shim.Error("Failed to update schedules count")
		}
	}
	up, err = t.updateList(stub, keySchedule + "_resIDs", resId)
	if err != nil {
		return shim.Error("Failed to udpate the resources list")
	}
	if up {
		countResources++
		err = stub.PutState(keySchedule + "_resources", itob(countResources))
		if err != nil {
			return shim.Error("Failed to update resources count")
		}
	}
	// It's success
	err = stub.SetEvent("eventInvoke", []byte{})
	if err != nil {
		return shim.Error(err.Error())
	}
	// Return this value in response
	return shim.Success(nil)
}
// true - new element added, false - list not changed
func (t *HeroesServiceChaincode) updateList(stub shim.ChaincodeStubInterface, listName string, newValue string) (bool, error) {
	inList, ls, err := t.findInList(stub, listName, newValue)
	if err != nil {
		return false, err
	}
	if inList {
		return false, nil
	}
	ls = append(ls, newValue)
	err = stub.PutState(listName, []byte(strings.Join(ls,";")))
	if err != nil {
		return false, err
	}
	return true, nil
}
func (t *HeroesServiceChaincode) selectFromList(stub shim.ChaincodeStubInterface, listName string) ([]string, error) {
	state, err := stub.GetState(listName)
	if err != nil {
		return []string{}, err
	}
	return strings.Split(string(state), ";"), nil
}
func (t *HeroesServiceChaincode) findInList(stub shim.ChaincodeStubInterface, listName string, val string) (bool, []string, error) {
	ls, err := t.selectFromList(stub, listName)
	if err != nil {
		return false, ls, err
	}
	for _, v := range ls {
		if v == val {
			return true, ls, nil
		}
	}
	return false, ls,  nil
}
func main() {
	// Start the chaincode and make it ready for future requests
	err := shim.Start(new(HeroesServiceChaincode))
	if err != nil {
		fmt.Printf("Error starting Heroes Service chaincode: %s", err)
	}
}